## Copyright 2016-2017 Jonas Hoersch (FIAS), Tom Brown (FIAS), Markus Schlott
## (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

import os
import pandas as pd
import numpy as np
import xarray as xr
import shutil
from six.moves import range
from contextlib import contextmanager
from tempfile import mkstemp
import logging
logger = logging.getLogger(__name__)

try:
    import cdsapi
    has_cdsapi = True
except ImportError:
    has_cdsapi = False

# Model and Projection Settings
projection = 'latlong'

@contextmanager
def _get_data(target=None, product='reanalysis-era5-single-levels', chunks=None, **updates):
    """Download ERA5 data from the Climate Data Store (CDS)"""

    if not has_cdsapi:
        raise RuntimeError(
            "Need installed cdsapi python package available from "
            "https://cds.climate.copernicus.eu/api-how-to"
        )

    # Default request
    request = {
        'product_type':'reanalysis',
        'format':'netcdf',
        'day':[
            '01','02','03','04','05','06','07','08','09','10','11','12',
            '13','14','15','16','17','18','19','20','21','22','23','24',
            '25','26','27','28','29','30','31'
        ],
        'time':[
            '00:00','01:00','02:00','03:00','04:00','05:00',
            '06:00','07:00','08:00','09:00','10:00','11:00',
            '12:00','13:00','14:00','15:00','16:00','17:00',
            '18:00','19:00','20:00','21:00','22:00','23:00'
        ],
        # 'area': [50, -1, 49, 1], # North, West, South, East. Default: global
        # 'grid': [0.25, 0.25], # Latitude/longitude grid: east-west (longitude) and north-south resolution (latitude). Default: 0.25 x 0.25
    }
    request.update(updates)

    assert {'year', 'month', 'variable'}.issubset(request), "Need to specify at least 'variable', 'year' and 'month'"

    result = cdsapi.Client().retrieve(
        product,
        request
    )

    if target is None:
        fd, target = mkstemp(suffix='.nc')
        os.close(fd)

    logger.info("Downloading request for {} variables to {}".format(len(request['variable']), target))

    result.download(target)

    with xr.open_dataset(target, chunks=chunks) as ds:
        yield ds

    os.unlink(target)

def _add_height(ds):
    """Convert geopotential 'z' to geopotential height following [1]

    References
    ----------
    [1] ERA5: surface elevation and orography, retrieved: 10.02.2019
    https://confluence.ecmwf.int/display/CKB/ERA5%3A+surface+elevation+and+orography

    """
    g0 = 9.80665
    z = ds['z']
    if 'time' in z.coords:
        z = z.isel(time=0, drop=True)
    ds['height'] = z/g0
    ds = ds.drop('z')
    return ds

def _area(xs, ys):
    # North, West, South, East. Default: global
    return [ys.start, xs.start, ys.stop, xs.stop]

def _rename_and_clean_coords(ds, add_lon_lat=True):
    """Rename 'longitude' and 'latitude' columns to 'x' and 'y'

    Optionally (add_lon_lat, default:True) preserves latitude and longitude columns as 'lat' and 'lon'.
    """

    ds = ds.rename({'longitude': 'x', 'latitude': 'y'})
    if add_lon_lat:
        ds = ds.assign_coords(lon=ds.coords['x'], lat=ds.coords['y'])
    return ds

def prepare_meta_era5(xs, ys, year, month, module):
    # Reference of the quantities
    # https://confluence.ecmwf.int/display/CKB/ERA5+data+documentation
    # Geopotential is aka Orography in the CDS:
    # https://confluence.ecmwf.int/pages/viewpage.action?pageId=78296105
    #
    # (shortName) | (name)                        | (paramId)
    # z           | Geopotential (CDS: Orography) | 129
    with _get_data(variable='orography',
                   year=year, month=month, day=1,
                   area=_area(xs, ys)) as ds:
        ds = _rename_and_clean_coords(ds)
        ds = _add_height(ds)

        t = pd.Timestamp(year=year, month=month, day=1)
        ds['time'] = pd.date_range(t, t + pd.DateOffset(months=1),
                                   freq='1h', closed='left')

        return ds.load()

def prepare_for_sarah(year, month, xs, ys, dx, dy, chunks=None):
    area = _area(xs, ys)
    grid = [dx, dy]

    with _get_data(area=area, grid=grid, year=year, month=month,
                   variable=['2m_temperature',
                             'toa_incident_solar_radiation',
                             'surface_solar_radiation_downwards',
                             'surface_net_solar_radiation'],
                   chunks=chunks) as ds:
        ds = _rename_and_clean_coords(ds, add_lon_lat=False)

        ds = ds.rename({'t2m': 'temperature'})
        ds = ds.rename({'tisr': 'influx_toa'})

        logger.debug("Calculating albedo")
        ds['albedo'] = (((ds['ssrd'] - ds['ssr'])/ds['ssrd']).fillna(0.)
                        .assign_attrs(units='(0 - 1)', long_name='Albedo'))
        ds = ds.drop(['ssrd', 'ssr'])

        logger.debug("Fixing units of influx_toa")
        # Convert from energy to power J m**-2 -> W m**-2
        ds['influx_toa'] /= 60.*60.
        ds['influx_toa'].attrs['units'] = 'W m**-2'

        logger.debug("Yielding ERA5 results")
        yield ds.chunk(chunks)
        logger.debug("Cleaning up ERA5")

def prepare_month_era5(year, month, xs, ys):
    area = _area(xs, ys)

    # Reference of the quantities
    # https://confluence.ecmwf.int/display/CKB/ERA5+data+documentation
    # (shortName) | (name)                                      | (paramId)
    # tisr        | TOA incident solar radiation                | 212
    # ssrd        | Surface Solar Rad Downwards                 | 169
    # ssr         | Surface net Solar Radiation                 | 176
    # fdir        | Total sky direct solar radiation at surface | 228021
    # ro          | Runoff                                      | 205
    # 2t          | 2 metre temperature                         | 167
    # sp          | Surface pressure                            | 134
    # stl4        | Soil temperature level 4                    | 236
    # fsr         | Forecast surface roughnes                   | 244

    with _get_data(area=area, year=year, month=month,
                   variable=[
                       '100m_u_component_of_wind',
                       '100m_v_component_of_wind',
                       '2m_temperature',
                       'runoff',
                       'soil_temperature_level_4',
                       'surface_net_solar_radiation',
                       'surface_pressure',
                       'surface_solar_radiation_downwards',
                       'toa_incident_solar_radiation',
                       'total_sky_direct_solar_radiation_at_surface'
                   ]) as ds, \
         _get_data(area=area, year=year, month=month, day=1,
                   variable=['forecast_surface_roughness', 'orography']) as ds_m:

        ds_m = ds_m.isel(time=0, drop=True)
        ds = xr.merge([ds, ds_m], join='left')

        ds = _rename_and_clean_coords(ds)
        ds = _add_height(ds)


        ds = ds.rename({'fdir': 'influx_direct', 'tisr': 'influx_toa'})
        with np.errstate(divide='ignore', invalid='ignore'):
            ds['albedo'] = (((ds['ssrd'] - ds['ssr'])/ds['ssrd']).fillna(0.)
                            .assign_attrs(units='(0 - 1)', long_name='Albedo'))
        ds['influx_diffuse'] = ((ds['ssrd'] - ds['influx_direct'])
                                .assign_attrs(units='J m**-2',
                                            long_name='Surface diffuse solar radiation downwards'))
        ds = ds.drop(['ssrd', 'ssr'])

        # Convert from energy to power J m**-2 -> W m**-2 and clip negative fluxes
        for a in ('influx_direct', 'influx_diffuse', 'influx_toa'):
            ds[a] = ds[a].clip(min=0.) / (60.*60.)
            ds[a].attrs['units'] = 'W m**-2'

        ds['wnd100m'] = (np.sqrt(ds['u100']**2 + ds['v100']**2)
                        .assign_attrs(units=ds['u100'].attrs['units'],
                                    long_name="100 metre wind speed"))
        ds = ds.drop(['u100', 'v100'])

        ds = ds.rename({'ro': 'runoff',
                        't2m': 'temperature',
                        'sp': 'pressure',
                        'stl4': 'soil temperature',
                        'fsr': 'roughness'
                        })

        ds['runoff'] = ds['runoff'].clip(min=0.)

        yield (year, month), ds

def tasks_monthly_era5(xs, ys, yearmonths, prepare_func, meta_attrs):
    if not isinstance(xs, slice):
        xs = slice(*xs.values[[0, -1]])
    if not isinstance(ys, slice):
        ys = slice(*ys.values[[0, -1]])

    return [dict(prepare_func=prepare_func, xs=xs, ys=ys, year=year, month=month)
            for year, month in yearmonths]

weather_data_config = {
    '_': dict(tasks_func=tasks_monthly_era5,
              prepare_func=prepare_month_era5)
}

meta_data_config = dict(prepare_func=prepare_meta_era5)



## Copyright 2016-2017 Gorm Andresen (Aarhus University), Jonas
## Hoersch (FIAS), Tom Brown (FIAS), Markus Schlott (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

from __future__ import absolute_import

import pandas as pd
import numpy as np
import xarray as xr
import pyproj
from six import iteritems
from itertools import groupby
from operator import itemgetter
import os
import glob

from ..config import cordex_dir
from ..gis import RotProj

# Model and Projection Settings
model = 'MPI-M-MPI-ESM-LR'
projection = RotProj(dict(proj='ob_tran', o_proj='latlong', lon_0=180,
                          o_lon_p=-162, o_lat_p=39.25))

def rename_and_clean_coords(ds):
    ds = ds.rename({'rlon': 'x', 'rlat': 'y'})
    # drop some coordinates and variables we do not use
    ds = ds.drop((set(ds.coords) | set(ds.data_vars))
                 & {'bnds', 'height', 'rotated_pole'})
    return ds

def prepare_data_cordex(fn, year, months, oldname, newname, xs, ys):
    with xr.open_dataset(fn) as ds:
        ds = rename_and_clean_coords(ds)
        ds = ds.rename({oldname: newname})
        ds = ds.sel(x=xs, y=ys)

        if newname in {'influx', 'outflux'}:
            # shift averaged data to beginning of bin
            ds = ds.assign_coords(time=(pd.to_datetime(ds.coords["time"].values)
                                        - pd.Timedelta(hours=1.5)))
        elif newname in {'runoff'}:
            # shift and fill 6hr average data to beginning of 3hr bins
            t = pd.to_datetime(ds.coords["time"].values)
            ds = ds.reindex(method='bfill', time=(t - pd.Timedelta(hours=3.)).union(t))

        for m in months:
            yield (year, m), ds.sel(time="{}-{}".format(year, m))

def prepare_static_data_cordex(fn, year, months, oldname, newname, xs, ys):
    with xr.open_dataset(fn) as ds:
        ds = rename_and_clean_coords(ds)
        ds = ds.rename({oldname: newname})
        ds = ds.sel(x=xs, y=ys)

        for m in months:
            yield (year, m), ds

def prepare_weather_types_cordex(fn, year, months, oldname, newname, xs, ys):
    with xr.open_dataset(fn) as ds:
        ds = ds.rename({oldname: newname})
        for m in months:
            yield (year, m), ds.sel(time="{}-{}".format(year, m))

def prepare_meta_cordex(xs, ys, year, month, template, height_config, module, model=model):
    fn = next(glob.iglob(template.format(year=year, model=model)))
    with xr.open_dataset(fn) as ds:
        ds = rename_and_clean_coords(ds)
        ds = ds.coords.to_dataset()
        meta = ds.sel(time="{}-{}".format(year, month),
                      x=xs, y=ys).load()

    xs = ds['x'].values
    ys = ds['y'].values

    height_config = height_config.copy()
    height_tasks_func = height_config.pop('tasks_func')
    height_task, = height_tasks_func(xs, ys, [(year, month)], meta_attrs={}, **height_config)
    height_prepare_func = height_task.pop('prepare_func')
    _, ds = height_prepare_func(**height_task)[0]

    meta['height'] = ds['height']

    return meta

def tasks_yearly_cordex(xs, ys, yearmonths, prepare_func, template, oldname, newname, meta_attrs):
    model = meta_attrs['model']

    if not isinstance(xs, slice):
        first, second, last = xs.values[[0,1,-1]]
        xs = slice(first - 0.1*(second - first), last + 0.1*(second - first))
    if not isinstance(ys, slice):
        first, second, last = ys.values[[0,1,-1]]
        ys = slice(first - 0.1*(second - first), last + 0.1*(second - first))

    return [dict(prepare_func=prepare_func,
                 xs=xs, ys=ys, oldname=oldname, newname=newname,
                 fn=next(glob.iglob(template.format(year=year, model=model))),
                 year=year, months=list(map(itemgetter(1), yearmonths)))
            for year, yearmonths in groupby(yearmonths, itemgetter(0))]

weather_data_config = {
    'influx': dict(tasks_func=tasks_yearly_cordex,
                   prepare_func=prepare_data_cordex,
                   oldname='rsds', newname='influx',
                   template=os.path.join(cordex_dir, '{model}', 'influx', 'rsds_*_{year}*.nc')),
    'outflux': dict(tasks_func=tasks_yearly_cordex,
                    prepare_func=prepare_data_cordex,
                    oldname='rsus', newname='outflux',
                    template=os.path.join(cordex_dir, '{model}', 'outflux', 'rsus_*_{year}*.nc')),
    'temperature': dict(tasks_func=tasks_yearly_cordex,
                        prepare_func=prepare_data_cordex,
                        oldname='tas', newname='temperature',
                        template=os.path.join(cordex_dir, '{model}', 'temperature', 'tas_*_{year}*.nc')),
    'humidity': dict(tasks_func=tasks_yearly_cordex,
                        prepare_func=prepare_data_cordex,
                        oldname='hurs', newname='humidity',
                        template=os.path.join(cordex_dir, '{model}', 'humidity', 'hurs_*_{year}*.nc')),
    'wnd10m': dict(tasks_func=tasks_yearly_cordex,
                   prepare_func=prepare_data_cordex,
                   oldname='sfcWind', newname='wnd10m',
                   template=os.path.join(cordex_dir, '{model}', 'wind', 'sfcWind_*_{year}*.nc')),
    'roughness': dict(tasks_func=tasks_yearly_cordex,
                      prepare_func=prepare_static_data_cordex,
                      oldname='rlst', newname='roughness',
                      template=os.path.join(cordex_dir, '{model}', 'roughness', 'rlst_*.nc')),
    'runoff': dict(tasks_func=tasks_yearly_cordex,
                   prepare_func=prepare_data_cordex,
                   oldname='mrro', newname='runoff',
                   template=os.path.join(cordex_dir, '{model}', 'runoff', 'mrro_*_{year}*.nc')),
    'height': dict(tasks_func=tasks_yearly_cordex,
                   prepare_func=prepare_static_data_cordex,
                   oldname='orog', newname='height',
                   template=os.path.join(cordex_dir, '{model}', 'altitude', 'orog_*.nc')),
    'CWT': dict(tasks_func=tasks_yearly_cordex,
                prepare_func=prepare_weather_types_cordex,
                oldname='CWT', newname='CWT',
                template=os.path.join(cordex_dir, '{model}', 'weather_types', 'CWT_*_{year}*.nc')),
}

meta_data_config = dict(prepare_func=prepare_meta_cordex,
                        template=os.path.join(cordex_dir, '{model}', 'temperature', 'tas_*_{year}*.nc'),
                        height_config=weather_data_config['height'])



## Copyright 2016-2017 Gorm Andresen (Aarhus University), Jonas Hoersch (FIAS), Tom Brown (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

from __future__ import absolute_import

import numpy as np
import pandas as pd
import xarray as xr
import scipy as sp, scipy.sparse
from collections import OrderedDict
from warnings import warn
from six import string_types, iteritems
from six.moves import map, range
from itertools import product
from functools import partial
import pyproj
from shapely.prepared import prep
from shapely.ops import transform
import rasterio as rio
import rasterio.warp
from rasterio.warp import Resampling

import logging
logger = logging.getLogger(__name__)

def spdiag(v):
    N = len(v)
    inds = np.arange(N+1, dtype=np.int32)
    return sp.sparse.csr_matrix((v, inds[:-1], inds), (N, N))

class RotProj(pyproj.Proj):
    def __call__(self, x, y, inverse=False, **kw):
        if inverse:
            gx, gy = super(RotProj, self).__call__(x, y,
                                                   inverse=False, **kw)
            return np.rad2deg(gx), np.rad2deg(gy)
        else:
            return super(RotProj, self).__call__(np.deg2rad(x),
                                                 np.deg2rad(y),
                                                 inverse=True, **kw)

def as_projection(p):
    if isinstance(p, pyproj.Proj):
        return p
    elif isinstance(p, string_types):
        return pyproj.Proj(dict(proj=p))
    else:
        return pyproj.Proj(p)

def reproject_shapes(shapes, p1, p2):
    """
    Project a collection of `shapes` from one projection `p1` to
    another projection `p2`

    Projections can be given as strings or instances of pyproj.Proj.
    Special care is taken for the case where the final projection is
    of type rotated pole as handled by RotProj.
    """

    if p1 == p2:
        return shapes

    if isinstance(p1, RotProj):
        if p2 == 'latlong':
            reproject_points = lambda x,y: p1(x, y, inverse=True)
        else:
            raise NotImplementedError("`p1` can only be a RotProj if `p2` is latlong!")

    if isinstance(p2, RotProj):
        shapes = reproject(shapes, p1, 'latlong')
        reproject_points = p2
    else:
        reproject_points = partial(pyproj.transform, as_projection(p1), as_projection(p2))

    def _reproject_shape(shape):
        return transform(reproject_points, shape)

    if isinstance(shapes, pd.Series):
        return shapes.map(_reproject_shape)
    elif isinstance(shapes, dict):
        return OrderedDict((k, _reproject_shape(v)) for k, v in iteritems(shapes))
    else:
        return list(map(_reproject_shape, shapes))

def reproject(shapes, p1, p2):
    warn("reproject has been renamed to reproject_shapes", DeprecationWarning)
    return reproject_shapes(shapes, p1, p2)
reproject.__doc__ = reproject_shapes.__doc__

def compute_indicatormatrix(orig, dest, orig_proj='latlong', dest_proj='latlong'):
    """
    Compute the indicatormatrix

    The indicatormatrix I[i,j] is a sparse representation of the ratio
    of the area in orig[j] lying in dest[i], where orig and dest are
    collections of polygons, i.e.

    A value of I[i,j] = 1 indicates that the shape orig[j] is fully
    contained in shape dest[j].

    Note that the polygons must be in the same crs.

    Parameters
    ---------
    orig : Collection of shapely polygons
    dest : Collection of shapely polygons

    Returns
    -------
    I : sp.sparse.lil_matrix
      Indicatormatrix
    """

    dest = reproject_shapes(dest, dest_proj, orig_proj)
    indicator = sp.sparse.lil_matrix((len(dest), len(orig)), dtype=np.float)

    try:
        from rtree.index import Index

        idx = Index()
        for j, o in enumerate(orig):
            idx.insert(j, o.bounds)

        for i, d in enumerate(dest):
            for j in idx.intersection(d.bounds):
                o = orig[j]
                area = d.intersection(o).area
                indicator[i,j] = area/o.area

    except ImportError:
        logger.warning("Rtree is not available. Falling back to slower algorithm.")

        dest_prepped = list(map(prep, dest))

        for i,j in product(range(len(dest)), range(len(orig))):
            if dest_prepped[i].intersects(orig[j]):
                area = dest[i].intersection(orig[j]).area
                indicator[i,j] = area/orig[j].area

    return indicator

def maybe_swap_spatial_dims(ds, namex='x', namey='y'):
    swaps = {}
    lx, rx = ds.indexes[namex][[0, -1]]
    uy, ly = ds.indexes[namey][[0, -1]]

    if lx > rx:
        swaps[namex] = slice(None, None, -1)
    if uy < ly:
        swaps[namey] = slice(None, None, -1)

    return ds.isel(**swaps) if swaps else ds

def _as_transform(x, y):
    lx, rx = x[[0, -1]]
    uy, ly = y[[0, -1]]

    dx = float(rx - lx)/float(len(x)-1)
    dy = float(uy - ly)/float(len(y)-1)

    return rio.transform.from_origin(lx, uy, dx, dy)

def regrid(ds, dimx, dimy, **kwargs):
    """
    Interpolate Dataset or DataArray `ds` to a new grid, using rasterio's
    reproject facility.

    See also: https://mapbox.github.io/rasterio/topics/resampling.html

    Parameters
    ----------
    ds : xr.Dataset|xr.DataArray
      N-dim data on a spatial grid
    dimx : pd.Index
      New x-coordinates in destination crs.
      dimx.name MUST refer to x-coord of ds.
    dimy : pd.Index
      New y-coordinates in destination crs.
      dimy.name MUST refer to y-coord of ds.
    **kwargs :
      Arguments passed to rio.wrap.reproject; of note:
      - resampling is one of gis.Resampling.{average,cubic,bilinear,nearest}
      - src_crs, dst_crs define the different crs (default: latlong)
    """
    namex = dimx.name
    namey = dimy.name

    ds = maybe_swap_spatial_dims(ds, namex, namey)

    src_transform = _as_transform(ds.indexes[namex],
                                  ds.indexes[namey])
    dst_transform = _as_transform(dimx, dimy)
    dst_shape = len(dimy), len(dimx)

    kwargs.update(dst_shape=dst_shape,
                  src_transform=src_transform,
                  dst_transform=dst_transform)
    kwargs.setdefault("src_crs", 'longlat')
    kwargs.setdefault("dst_crs", 'longlat')

    def _reproject(src, dst_shape, **kwargs):
        dst = np.empty(src.shape[:-2] + dst_shape, dtype=src.dtype)
        rio.warp.reproject(np.asarray(src), dst, **kwargs)
        return dst

    data_vars = ds.data_vars.values() if isinstance(ds, xr.Dataset) else (ds,)
    dtypes = {da.dtype for da in data_vars}
    assert len(dtypes) == 1, "regrid can only reproject datasets with homogeneous dtype"

    return (
        xr.apply_ufunc(_reproject, ds,
                       input_core_dims=[[namey, namex]],
                       output_core_dims=[['yout', 'xout']],
                       output_dtypes=[dtypes.pop()],
                       output_sizes={'yout': dst_shape[0], 'xout': dst_shape[1]},
                       dask='parallelized',
                       kwargs=kwargs)
        .rename({'yout': namey, 'xout': namex})
        .assign_coords(**{namey: (namey, dimy, ds.coords[namey].attrs),
                            namex: (namex, dimx, ds.coords[namex].attrs)})
        .assign_attrs(**ds.attrs)
    )



# -*- coding: utf-8 -*-

## Copyright 2016-2017 Gorm Andresen (Aarhus University), Jonas Hoersch (FIAS), Tom Brown (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

from __future__ import absolute_import

import os
import yaml
from six import string_types
from operator import itemgetter
import numpy as np
from scipy.signal import fftconvolve
from pkg_resources import resource_stream

import logging
logger = logging.getLogger(name=__name__)

def get_windturbineconfig(turbine):
    """Load the 'turbine'.yaml file from local disk and provide a turbine dict."""

    res_name = "resources/windturbine/" + turbine + ".yaml"
    turbineconf = yaml.safe_load(resource_stream(__name__, res_name))
    V, POW, hub_height = itemgetter('V', 'POW', 'HUB_HEIGHT')(turbineconf)
    return dict(V=np.array(V), POW=np.array(POW), hub_height=hub_height, P=np.max(POW))

def get_solarpanelconfig(panel):
    res_name = "resources/solarpanel/" + panel + ".yaml"
    return yaml.safe_load(resource_stream(__name__, res_name))

def solarpanel_rated_capacity_per_unit(panel):
    # unit is m^2 here

    if isinstance(panel, string_types):
        panel = get_solarpanelconfig(panel)

    model = panel.get('model', 'huld')
    if model == 'huld':
        return panel['efficiency']
    elif model == 'bofinger':
        # one unit in the capacity layout is interpreted as one panel of a
        # capacity (A + 1000 * B + log(1000) * C) * 1000W/m^2 * (k / 1000)
        A, B, C = itemgetter('A', 'B', 'C')(panel)
        return (A + B * 1000. + C * np.log(1000.))*1e3

def windturbine_rated_capacity_per_unit(turbine):
    if isinstance(turbine, string_types):
        turbine = get_windturbineconfig(turbine)

    return turbine['P']

def windturbine_smooth(turbine, params={}):
    '''
    Smooth the powercurve in `turbine` with a gaussian kernel

    Parameters
    ----------
    turbine : dict
        Turbine config with at least V and POW
    params : dict
        Allows adjusting fleet availability eta, mean Delta_v and
        stdev sigma. Defaults to values from Andresen's paper: 0.95,
        1.27 and 2.29, respectively.

    Returns
    -------
    turbine : dict
        Turbine config with a smoothed power curve

    References
    ----------
    G. B. Andresen, A. A. Søndergaard, M. Greiner, Validation of
    Danish wind time series from a new global renewable energy atlas
    for energy system analysis, Energy 93, Part 1 (2015) 1074–1088.
    '''

    if not isinstance(params, dict):
        params = {}

    eta = params.setdefault('eta', 0.95)
    Delta_v = params.setdefault('Delta_v', 1.27)
    sigma = params.setdefault('sigma', 2.29)

    def kernel(v_0):
        # all velocities in m/s
        return (1./np.sqrt(2*np.pi*sigma*sigma) *
                np.exp(-(v_0 - Delta_v)*(v_0 - Delta_v)/(2*sigma*sigma) ))

    def smooth(velocities, power):
        # interpolate kernel and power curve to the same, regular velocity grid
        velocities_reg = np.linspace(-50., 50., 1001)
        power_reg = np.interp(velocities_reg, velocities, power)
        kernel_reg = kernel(velocities_reg)

        # convolve power and kernel
        # the downscaling is necessary because scipy expects the velocity
        # increments to be 1., but here, they are 0.1
        convolution = 0.1*fftconvolve(power_reg, kernel_reg, mode='same')

        # sample down so power curve doesn't get too long
        velocities_new = np.linspace(0., 35., 72)
        power_new = eta * np.interp(velocities_new, velocities_reg, convolution)

        return velocities_new, power_new

    turbine = turbine.copy()
    turbine['V'], turbine['POW'] = smooth(turbine['V'], turbine['POW'])
    turbine['P'] = np.max(turbine['POW'])

    if any(turbine['POW'][np.where(turbine['V'] == 0.0)] > 1e-2):
        logger.warn("Oversmoothing detected with parameters {p}. " +
                    "Turbine generates energy at 0 m/s wind speeds".format(p=params))

    return turbine



## Copyright 2016-2017 Gorm Andresen (Aarhus University), Jonas Hoersch (FIAS), Tom Brown (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

from __future__ import absolute_import

import xarray as xr
import numpy as np
import os, sys
from six import string_types

import logging
logger = logging.getLogger(__name__)

from . import config, datasets

from .convert import (convert_and_aggregate, heat_demand, hydro, temperature,
                      wind, pv, runoff, solar_thermal, soil_temperature)
from .preparation import (cutout_do_task, cutout_prepare,
                          cutout_produce_specific_dataseries,
                          cutout_get_meta, cutout_get_meta_view)
from .gis import compute_indicatormatrix

class Cutout(object):
    def __init__(self, name=None, cutout_dir=config.cutout_dir, **cutoutparams):
        self.name = name

        self.cutout_dir = os.path.join(cutout_dir, name)
        self.prepared = False

        if 'bounds' in cutoutparams:
            x1, y1, x2, y2 = cutoutparams.pop('bounds')
            cutoutparams.update(xs=slice(x1, x2),
                                ys=slice(y2, y1))

        if os.path.isdir(self.cutout_dir):
            self.meta = meta = xr.open_dataset(self.datasetfn()).stack(**{'year-month': ('year', 'month')})
            # check datasets very rudimentarily, series and coordinates should be checked as well
            if all(os.path.isfile(self.datasetfn(ym)) for ym in meta.coords['year-month'].to_index()):
                self.prepared = True
            else:
                assert False

            if 'module' in meta.attrs:
                cutoutparams['module'] = meta.attrs['module']
            else:
                logger.warning('module not given in meta file of cutout, assuming it is NCEP')
                cutoutparams['module'] = 'ncep'

            if {"xs", "ys", "years", "months"}.intersection(cutoutparams):
                # Assuming the user is interested in a subview into
                # the data, update meta in place for the time
                # dimension and save the xs, ys slices, separately
                self.meta = meta = self.get_meta_view(**cutoutparams)
                logger.info("Assuming a view into the prepared cutout: %s", self)

        else:
            logger.info("Cutout %s not found in directory %s, building new one", name, cutout_dir)

            if 'module' not in cutoutparams:
                d = config.weather_dataset.copy()
                d.update(cutoutparams)
                cutoutparams = d

        self.dataset_module = sys.modules['atlite.datasets.' + cutoutparams['module']]

        if not self.prepared:
            if {"xs", "ys", "years"}.difference(cutoutparams):
                raise TypeError("Arguments `xs`, `ys` and `years` need to be specified")
            self.meta = self.get_meta(**cutoutparams)

    def datasetfn(self, *args):
        dataset = None

        if len(args) == 2:
            dataset = args
        elif len(args) == 1:
            dataset = args[0]
        else:
            dataset = None
        return os.path.join(self.cutout_dir, ("meta.nc"
                                              if dataset is None
                                              else "{}{:0>2}.nc".format(*dataset)))

    @property
    def meta_data_config(self):
        return self.dataset_module.meta_data_config

    @property
    def weather_data_config(self):
        return self.dataset_module.weather_data_config

    @property
    def projection(self):
        return self.dataset_module.projection

    @property
    def coords(self):
        return self.meta.coords

    @property
    def shape(self):
        return len(self.coords["y"]), len(self.coords["x"])

    @property
    def extent(self):
        return (list(self.coords["x"].values[[0, -1]]) +
                list(self.coords["y"].values[[-1, 0]]))


    def grid_coordinates(self):
        xs, ys = np.meshgrid(self.coords["x"], self.coords["y"])
        return np.asarray((np.ravel(xs), np.ravel(ys))).T

    def grid_cells(self):
        from shapely.geometry import box
        coords = self.grid_coordinates()
        span = (coords[self.shape[1]+1] - coords[0]) / 2
        return [box(*c) for c in np.hstack((coords - span, coords + span))]

    def __repr__(self):
        yearmonths = self.coords['year-month'].to_index()
        return ('<Cutout {} x={:.2f}-{:.2f} y={:.2f}-{:.2f} time={}/{}-{}/{} {}prepared>'
                .format(self.name,
                        self.coords['x'].values[0], self.coords['x'].values[-1],
                        self.coords['y'].values[0], self.coords['y'].values[-1],
                        yearmonths[0][0],  yearmonths[0][1],
                        yearmonths[-1][0], yearmonths[-1][1],
                        "" if self.prepared else "UN"))

    def indicatormatrix(self, shapes, shapes_proj='latlong'):
        return compute_indicatormatrix(self.grid_cells(), shapes, self.projection, shapes_proj)

    ## Preparation functions

    get_meta = cutout_get_meta

    get_meta_view = cutout_get_meta_view

    prepare = cutout_prepare

    produce_specific_dataseries = cutout_produce_specific_dataseries

    ## Conversion and aggregation functions

    convert_and_aggregate = convert_and_aggregate

    heat_demand = heat_demand

    temperature = temperature

    soil_temperature = soil_temperature

    solar_thermal = solar_thermal

    wind = wind

    pv = pv

    runoff = runoff

    hydro = hydro



## Copyright 2016-2017 Jonas Hoersch (FIAS), Tom Brown (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

from __future__ import absolute_import

import pandas as pd
import numpy as np
import xarray as xr
from functools import partial
import pyproj
from six import iteritems
from itertools import groupby
from operator import itemgetter
import os
import glob
import logging
logger = logging.getLogger(__name__)

from collections import deque
from contextlib import contextmanager

@contextmanager
def receive(it):
    yield next(it)
    for i in it: pass

def as_slice(zs, pad=True):
    if not isinstance(zs, slice):
        first, second, last = np.asarray(zs)[[0,1,-1]]
        dz = 0.1 * (second - first) if pad else 0.
        zs = slice(first - dz, last + dz)
    return zs

from ..config import sarah_dir
from ..gis import regrid, Resampling, maybe_swap_spatial_dims
from .era5 import prepare_for_sarah

# Model and Projection Settings
projection = 'latlong'
resolution = None

def _rename_and_clean_coords(ds, add_lon_lat=True):
    ds = ds.rename({'lon': 'x', 'lat': 'y'})
    ds = maybe_swap_spatial_dims(ds)
    if add_lon_lat:
        ds = ds.assign_coords(lon=ds.coords['x'], lat=ds.coords['y'])
    return ds

def prepare_meta_sarah(xs, ys, year, month, template_sis, template_sid, module, resolution=resolution):
    fns = [next(glob.iglob(t.format(year=year, month=month)))
           for t in (template_sis, template_sid)]

    with xr.open_mfdataset(fns, compat='identical') as ds:
        ds = _rename_and_clean_coords(ds)
        ds = ds.coords.to_dataset()

        t = pd.Timestamp(year=year, month=month, day=1)
        ds['time'] = pd.date_range(t, t + pd.DateOffset(months=1),
                                   freq='1h', closed='left')

        if resolution is not None:
            def p(s):
                s += 0.1*resolution
                return s - (s % resolution)
            xs = np.arange(p(xs.start), p(xs.stop) + 1.1*resolution, resolution)
            ys = np.arange(p(ys.start), p(ys.stop) - 0.1*resolution, - resolution)
            ds = ds.sel(x=xs, y=ys, method='nearest')
        else:
            ds = ds.sel(x=as_slice(xs), y=as_slice(ys))

        return ds.load()

def prepare_month_sarah(era5_func, xs, ys, year, month, template_sis, template_sid, resolution):
    with xr.open_mfdataset(template_sis.format(year=year, month=month)) as ds_sis, \
         xr.open_mfdataset(template_sid.format(year=year, month=month)) as ds_sid:
        ds = xr.merge([ds_sis, ds_sid])

        ds = _rename_and_clean_coords(ds, add_lon_lat=False)
        ds = ds.sel(x=as_slice(xs), y=as_slice(ys))

        def interpolate(ds, dim='time'):
            def _interpolate1d(y):
                nan = np.isnan(y)
                if nan.all(): return y
                x = lambda z: z.nonzero()[0]
                y[nan] = np.interp(x(nan), x(~nan), y[~nan])
                return y

            def _interpolate(a):
                return a.map_blocks(partial(np.apply_along_axis, _interpolate1d, -1), dtype=a.dtype)

            data_vars = ds.data_vars.values() if isinstance(ds, xr.Dataset) else (ds,)
            dtypes = {da.dtype for da in data_vars}
            assert len(dtypes) == 1, "interpolate only supports datasets with homogeneous dtype"

            return xr.apply_ufunc(_interpolate, ds,
                                input_core_dims=[[dim]],
                                output_core_dims=[[dim]],
                                output_dtypes=[dtypes.pop()],
                                output_sizes={dim: len(ds.indexes[dim])},
                                dask='allowed',
                                keep_attrs=True)

        ds = interpolate(ds)

        def hourly_mean(ds):
            ds1 = ds.isel(time=slice(None, None, 2))
            ds2 = ds.isel(time=slice(1, None, 2))
            ds2 = ds2.assign_coords(time=ds2.indexes['time'] - pd.Timedelta(30, 'm'))
            ds = ((ds1 + ds2)/2)
            ds.attrs = ds1.attrs
            for v in ds.variables:
                ds[v].attrs = ds1[v].attrs
            return ds

        ds = hourly_mean(ds)

        ds['influx_diffuse'] = ((ds['SIS'] - ds['SID'])
                                .assign_attrs(long_name='Surface Diffuse Shortwave Flux',
                                              units='W m-2'))
        ds = ds.rename({'SID': 'influx_direct'}).drop('SIS')

        if resolution is not None:
            ds = regrid(ds, xs, ys, resampling=Resampling.average)

        x = ds.indexes['x']
        y = ds.indexes['y']
        xs = slice(*x[[0,-1]])
        ys = slice(*y[[0,-1]])

        lx, rx = x[[0, -1]]
        uy, ly = y[[0, -1]]
        dx = float(rx - lx)/float(len(x)-1)
        dy = float(uy - ly)/float(len(y)-1)

        logger.debug("Getting ERA5 data")
        with receive(era5_func(year, month, xs, ys, dx, dy, chunks=dict(time=24))) as ds_era:
            logger.debug("Merging SARAH and ERA5 data")
            ds_era = ds_era.assign_coords(x=ds.indexes['x'], y=ds.indexes['y'])
            ds = xr.merge([ds, ds_era]).assign_attrs(ds.attrs)
            ds = ds.assign_coords(lon=ds.coords['x'], lat=ds.coords['y'])

            yield ((year, month), ds)

def tasks_monthly_sarah(xs, ys, yearmonths, prepare_func, era5_func, template_sis, template_sid, meta_attrs):
    resolution = meta_attrs.get('resolution', None)

    return [dict(prepare_func=prepare_func,
                 era5_func=era5_func,
                 template_sis=template_sis, template_sid=template_sid,
                 xs=xs, ys=ys, year=year, month=month,
                 resolution=resolution)
            for year, month in yearmonths]

weather_data_config = {
    '_': dict(tasks_func=tasks_monthly_sarah,
              prepare_func=prepare_month_sarah,
              era5_func=prepare_for_sarah,
              template_sid=os.path.join(sarah_dir, 'sid', 'SIDin{year}{month:02}*.nc'),
              template_sis=os.path.join(sarah_dir, 'sis', 'SISin{year}{month:02}*.nc'))
}

meta_data_config = dict(prepare_func=prepare_meta_sarah,
                        template_sid=weather_data_config['_']['template_sid'],
                        template_sis=weather_data_config['_']['template_sis'])




# This script uses the FIAS vresutils to built heat demand for each
# European country based on its population distribution (at NUTS 3
# resolution)


import sys, os

from vresutils import shapes as vshapes, mapping as vmapping, transfer as vtransfer
from load import europe

import pandas as pd
import numpy as np

import scipy as sp

import atlite

cutout = atlite.Cutout('europe-2011-2014')

#list of grid cells
grid_cells = cutout.grid_cells()
print(len(grid_cells))


#pd.Series nuts3 code -> 2-letter country codes
mapping = vmapping.countries_to_nuts3()

countries = mapping.value_counts().index.sort_values()


#pd.Series
gdp,pop = europe.gdppop_nuts3()

#pd.Series nuts3 code -> polygon
nuts3 = pd.Series(vshapes.nuts3(tolerance=None, minarea=0.))

#2014 populations

pop["CH040"] = pop["CH04"]
pop["CH070"] = pop["CH07"]
#Montenegro
pop["ME000"] = 650
pop["AL1"] = 2893
pop["BA1"] = 3871
pop["RS1"] = 7210

pop[nuts3.index][pd.isnull(pop[nuts3.index])]

pop.groupby(mapping).sum()

#takes 10 minutes
pop_map = pd.DataFrame()

for country in countries:
    print(country)
    country_nuts = mapping.index[mapping == country]
    trans_matrix = cutout.indicatormatrix(np.asarray(nuts3[country_nuts]))
    #CH has missing bits
    country_pop = pop[country_nuts].fillna(0.)
    pop_map[country] = np.array(trans_matrix.multiply(np.asarray(country_pop)).sum(axis=1))[:,0]

print(pop_map.sum())

#%%time
pop_matrix = sp.sparse.csr_matrix(pop_map.T)
index = pd.Index(countries,name="countries")

#%%time
hd = cutout.heat_demand(matrix=pop_matrix, index=index)

hd.T.to_pandas().to_pickle("heat_demand.pkl")



## Copyright 2016-2017 Gorm Andresen (Aarhus University), Jonas Hoersch (FIAS), Tom Brown (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

from __future__ import absolute_import

import xarray as xr

def aggregate_sum(da):
    return da.sum('time')

def aggregate_matrix(da, matrix, index):
    da = da.stack(spatial=('y', 'x')).transpose('spatial', 'time')
    return xr.DataArray(matrix * da,
                        [index, da.coords['time']])



# -*- coding: utf-8 -*-

## Copyright 2016-2017 Jonas Hoersch (RLI)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""
import xarray as xr
import geopandas as gpd
import pandas as pd
import numpy as np

import scipy.sparse as spp
from collections import namedtuple
from shapely.geometry import Point
from six import string_types

from .utils import make_optional_progressbar

import logging
logger = logging.getLogger(__name__)

Basins = namedtuple('Basins', ['plants', 'meta', 'shapes'])

def find_basin(shapes, lon, lat):
    hids = shapes.index[shapes.intersects(Point(lon, lat))]
    if len(hids) > 1:
        logger.warning(f"The point ({lon}, {lat}) is in several basins: {hids}. "
                        "Assuming the first one.")
    return hids[0]

def find_upstream_basins(meta, hid):
    hids = [hid]
    i = 0
    while i < len(hids):
        hids.extend(meta.index[meta['NEXT_DOWN'] == hids[i]])
        i += 1
    return hids

def determine_basins(plants, hydrobasins, show_progress=True):
    if isinstance(hydrobasins, string_types):
        hydrobasins = gpd.read_file(hydrobasins)

    assert isinstance(hydrobasins, gpd.GeoDataFrame), (
        "hydrobasins should be passed as a filename or a GeoDataFrame, "
        "but received `type(hydrobasins) = {}`".format(type(hydrobasins))
    )

    missing_columns = (pd.Index(['HYBAS_ID', 'DIST_MAIN', 'NEXT_DOWN', 'geometry'])
                       .difference(hydrobasins.columns))
    assert missing_columns.empty, (
        "Couldn't find the column(s) {} in the hydrobasins dataset."
        .format(", ".join(missing_columns))
    )

    hydrobasins = hydrobasins.set_index("HYBAS_ID")

    meta = hydrobasins[hydrobasins.columns.difference(('geometry',))]
    shapes = hydrobasins['geometry']

    maybe_progressbar = make_optional_progressbar(
        show_progress, "Determine upstream basins per plant", len(plants)
    )

    plant_basins = []
    for p in maybe_progressbar(plants.itertuples()):
        hid = find_basin(shapes, p.lon, p.lat)
        plant_basins.append((hid, find_upstream_basins(meta, hid)))
    plant_basins = pd.DataFrame(plant_basins, columns=['hid', 'upstream'], index=plants.index)

    unique_basins = pd.Index(plant_basins['upstream'].sum()).unique().rename("hid")
    return Basins(plant_basins, meta.loc[unique_basins], shapes.loc[unique_basins])

def shift_and_aggregate_runoff_for_plants(basins, runoff, flowspeed=1, show_progress=True):
    inflow = xr.DataArray(np.zeros((len(basins.plants), runoff.indexes["time"].size)),
                          [('plant', basins.plants.index),
                           ('time' , runoff.coords["time"])])

    maybe_progressbar = make_optional_progressbar(
        show_progress, "Shift and aggregate runoff by plant", len(basins.plants)
    )

    for ppl in maybe_progressbar(basins.plants.itertuples()):
        inflow_plant = inflow.loc[dict(plant=ppl.Index)]
        distances = (basins.meta.loc[ppl.upstream, "DIST_MAIN"] -
                     basins.meta.at[ppl.hid, "DIST_MAIN"])
        nhours = (distances * (flowspeed * 3.6) + 0.5).astype(int)

        for b in ppl.upstream:
            inflow_plant += runoff.sel(hid=b).shift(time=nhours.at[b])

    return inflow



# -*- coding: utf-8 -*-

## Copyright 2016-2017 Gorm Andresen (Aarhus University), Jonas Hoersch (FIAS), Tom Brown (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

import progressbar as pgb

def make_optional_progressbar(show, prefix, max_value):
    if show:
        widgets = [
            pgb.widgets.Percentage(),
            ' ', pgb.widgets.SimpleProgress(format='(%s)' % pgb.widgets.SimpleProgress.DEFAULT_FORMAT),
            ' ', pgb.widgets.Bar(),
            ' ', pgb.widgets.Timer(),
            ' ', pgb.widgets.ETA()
        ]
        if not prefix.endswith(": "):
            prefix = prefix.strip() + ": "
        maybe_progressbar = pgb.ProgressBar(prefix=prefix, widgets=widgets, max_value=max_value)
    else:
        maybe_progressbar = lambda x: x

    return maybe_progressbar




# This script creates the heat demand for a city

# It is assumed the city is half on grid cell 43 and half on grid cell
# 44

# It assumes you have already created a cutout for Europe called
# "europe-2011-2014" in the directory "/home/vres/data/cutouts"



from scipy import sparse


import atlite


cutout = atlite.Cutout('europe-2011-2014')

# A sparse matrix describing the city relative to the grid coordinates
matrix = sparse.csr_matrix(([0.5,0.5], ([0,1],[43,44])), shape=(2, len(cutout.grid_coordinates())))

print(matrix)

hd = cutout.heat_demand(matrix)

print(hd)




# In this example we assume you have set in config.py

# ncep_dir = '/path/to/weather_data/'

# where the files have format e.g.

# 'ncep_dir/{year}{month:0>2}/tmp2m.*.grb2'

import logging
logging.basicConfig(level=logging.DEBUG)

import atlite

cutout = atlite.Cutout(name="europe-2011-01",
                       module="ncep",
                       xs=slice(-12.18798349, 41.56244222),
                       ys=slice(71.65648314, 33.56459975),
                       years=slice(2011, 2011),
                       months=slice(1,1))

#this is where all the work happens - it took 105 minutes on FIAS'
#beast resi, with 16 cores; the resulting cutout takes 57 GB
cutout.prepare()



#Dataset Settings
gebco_path = '/home/vres-climate/data/GEBCO_2014_2D.nc'
cutout_dir = '/home/vres/data/cutouts'
ncep_dir = '/home/vres-climate/data/rda_ucar'
cordex_dir = '/home/vres-climate/data/cordex/RCP8.5'
sarah_dir = '/home/vres-climate/data/sarah_v2'
weather_dataset = {'module': 'cordex', 'model': 'MPI-M-MPI-ESM-LR'}



## Copyright 2016-2017 Gorm Andresen (Aarhus University), Jonas Hoersch (FIAS), Tom Brown (FIAS)

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Renewable Energy Atlas Lite (Atlite)

Light-weight version of Aarhus RE Atlas for converting weather data to power systems data
"""

from __future__ import absolute_import

import xarray as xr
import dask
import pandas as pd
import numpy as np
import os, shutil
import logging
import tempfile
import shutil
import subprocess
from glob import glob
from six import itervalues
from six.moves import map
from multiprocessing import Pool

logger = logging.getLogger(__name__)

def cutout_do_task(task, write_to_file=True):
    task = task.copy()
    prepare_func = task.pop('prepare_func')
    if write_to_file:
        datasetfns = task.pop('datasetfns')

    # Force dask to use just one thread (to save memory)
    with dask.config.set(scheduler='single-threaded'):
        try:
            data = prepare_func(**task)
            if data is None:
                data = []

            if write_to_file:
                for yearmonth, ds in data:
                    ## TODO : rewrite using plain netcdf4 to add variables
                    ## to the same file one by one
                    fn = datasetfns[yearmonth]
                    logger.debug("Writing to %s", os.path.basename(fn))
                    ds.to_netcdf(fn)
                    logger.debug("Write variable(s) %s to %s generated by %s",
                                ", ".join(ds.data_vars),
                                os.path.basename(fn),
                                prepare_func.__name__)
            else:
                return data
        except Exception as e:
            logger.exception("Exception occured in the task with prepare_func `%s`: %s",
                            prepare_func.__name__, e.args[0])
            raise e

def cutout_prepare(cutout, overwrite=False, nprocesses=None, gebco_height=False):
    if cutout.prepared and not overwrite:
        raise ArgumentError("The cutout is already prepared. If you want to recalculate it, "
                            "anyway, then you must supply an `overwrite=True` argument.")

    logger.info("Starting preparation of cutout '%s'", cutout.name)

    cutout_dir = cutout.cutout_dir
    yearmonths = cutout.coords['year-month'].to_index()
    xs = cutout.meta.indexes['x']
    ys = cutout.meta.indexes['y']

    if gebco_height:
        logger.info("Interpolating gebco to the dataset grid")
        cutout.meta['height'] = _prepare_gebco_height(xs, ys)

    # Delete cutout_dir
    if os.path.isdir(cutout_dir):
        logger.debug("Deleting cutout_dir '%s'", cutout_dir)
        shutil.rmtree(cutout_dir)

    os.mkdir(cutout_dir)
    cutout.meta.unstack('year-month').to_netcdf(cutout.datasetfn())

    # Compute data and fill files
    tasks = []
    for series in itervalues(cutout.weather_data_config):
        series = series.copy()
        series['meta_attrs'] = cutout.meta.attrs
        tasks_func = series.pop('tasks_func')
        tasks += tasks_func(xs=xs, ys=ys, yearmonths=yearmonths, **series)
    for i, t in enumerate(tasks):
        def datasetfn_with_id(ym):
            base, ext = os.path.splitext(cutout.datasetfn(ym))
            return base + "-{}".format(i) + ext
        t['datasetfns'] = {ym: datasetfn_with_id(ym) for ym in yearmonths.tolist()}

    logger.info("%d tasks have been collected. Starting running them on %s.",
                len(tasks),
                ("%d processes" % nprocesses)
                if nprocesses is not None
                else "all processors")

    pool = Pool(processes=nprocesses)
    try:
        pool.map(cutout_do_task, tasks)
    except Exception as e:
        pool.terminate()
        logger.info("Preparation of cutout '%s' has been interrupted by an exception. "
                    "Purging the incomplete cutout_dir.",
                    cutout.name)
        shutil.rmtree(cutout_dir)
        raise e
    pool.close()

    logger.info("Merging variables into monthly compound files")

    for fn in map(cutout.datasetfn, yearmonths.tolist()):
        base, ext = os.path.splitext(fn)
        fns = glob(base + "-*" + ext)
        if len(fns) == 1 and not gebco_height:
            # Fast-path
            os.rename(fns[0], fn)
        else:
            with xr.open_mfdataset(fns) as ds:
                if gebco_height:
                    ds['height'] = cutout.meta['height']

                ds.to_netcdf(fn)

            for tfn in fns: os.unlink(tfn)
        logger.debug("Completed file %s", os.path.basename(fn))

    logger.info("Cutout '%s' has been successfully prepared", cutout.name)
    cutout.prepared = True

def cutout_produce_specific_dataseries(cutout, yearmonth, series_name):
    xs = cutout.coords['x']
    ys = cutout.coords['y']
    series = cutout.weather_data_config[series_name].copy()
    series['meta_attrs'] = cutout.meta.attrs
    tasks_func = series.pop('tasks_func')
    tasks = tasks_func(xs=xs, ys=ys, yearmonths=[yearmonth], **series)

    assert len(tasks) == 1
    data = cutout_do_task(tasks[0], write_to_file=False)
    assert len(data) == 1 and data[0][0] == yearmonth
    return data[0][1]

def cutout_get_meta(cutout, xs, ys, years, months=None, **dataset_params):
    if months is None:
        months = slice(1, 12)
    if ys.stop > ys.start:
        logger.warn("ys slices are expected from north to south, i.e. slice(70, 40) for europe.")
        ys = slice(ys.stop, ys.start, -ys.step if ys.step is not None else None)

    meta_kwds = cutout.meta_data_config.copy()
    meta_kwds.update(dataset_params)

    prepare_func = meta_kwds.pop('prepare_func')
    ds = prepare_func(xs=xs, ys=ys, year=years.stop, month=months.stop, **meta_kwds)
    ds.attrs.update(dataset_params)

    start, second, end = map(pd.Timestamp, ds.coords['time'].values[[0, 1, -1]])
    month_start = pd.Timestamp("{}-{}".format(years.stop, months.stop))

    offset_start = (start - month_start)
    offset_end = (end - (month_start + pd.offsets.MonthBegin()))
    step = (second - start).components.hours

    ds.coords["time"] = pd.date_range(
        start=pd.Timestamp("{}-{}".format(years.start, months.start)) + offset_start,
        end=(month_start + pd.offsets.MonthBegin() + offset_end),
        freq='h' if step == 1 else ('%dh' % step))

    ds.coords["year"] = range(years.start, years.stop+1)
    ds.coords["month"] = range(months.start, months.stop+1)
    ds = ds.stack(**{'year-month': ('year', 'month')})

    return ds

def cutout_get_meta_view(cutout, xs=None, ys=None, years=slice(None), months=slice(None), **dataset_params):
    meta = cutout.meta

    if xs is not None:
        meta.attrs.setdefault('view', {})['x'] = xs
    if ys is not None:
        meta.attrs.setdefault('view', {})['y'] = ys

    meta = (meta
            .unstack('year-month')
            .sel(year=years, month=months, **meta.attrs.get('view', {}))
            .stack(**{'year-month': ('year', 'month')}))

    meta = meta.sel(time=slice(*("{:04}-{:02}".format(*ym)
                                 for ym in meta['year-month'][[0,-1]].to_index())))

    return meta


def _prepare_gebco_height(xs, ys, gebco_fn=None):
    # gebco bathymetry heights for underwater
    if gebco_fn is None:
        from .config import gebco_path
        gebco_fn = gebco_path

    tmpdir = tempfile.mkdtemp()
    cornersc = np.array(((xs[0], ys[0]), (xs[-1], ys[-1])))
    minc = np.minimum(*cornersc)
    maxc = np.maximum(*cornersc)
    span = (maxc - minc)/(np.asarray((len(xs), len(ys)))-1)
    minx, miny = minc - span/2.
    maxx, maxy = maxc + span/2.

    tmpfn = os.path.join(tmpdir, 'resampled.nc')
    try:
        ret = subprocess.call(['gdalwarp', '-of', 'NETCDF',
                               '-ts', str(len(xs)), str(len(ys)),
                               '-te', str(minx), str(miny), str(maxx), str(maxy),
                               '-r', 'average',
                               gebco_fn, tmpfn])
        assert ret == 0, "gdalwarp was not able to resample gebco"
    except OSError:
        logger.warning("gdalwarp was not found for resampling gebco. "
                       "Next-neighbour interpolation will be used instead!")
        tmpfn = gebco_path

    with xr.open_dataset(tmpfn) as ds_gebco:
        height = (ds_gebco.rename({'lon': 'x', 'lat': 'y', 'Band1': 'height'})
                          .reindex(x=xs, y=ys, method='nearest')
                          .load()['height'])
    shutil.rmtree(tmpdir)
    return height



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import xarray as xr

def SolarPosition(ds):
    """
    Compute solar azimuth and altitude

    Solar altitude errors are up to 1.5 deg during sun-rise and set, but at
    0.05-0.1 deg during daytime.

    References
    ----------
    [1] Michalsky, J. J., The astronomical almanac’s algorithm for approximate
    solar position (1950–2050), Solar Energy, 40(3), 227–235 (1988).
    [2] Sproul, A. B., Derivation of the solar geometric relationships using
    vector analysis, Renewable Energy, 32(7), 1187–1205 (2007).
    [3] Kalogirou, Solar Energy Engineering (2009).

    More accurate algorithms would be
    ---------------------------------
    [4] I. Reda and A. Andreas, Solar position algorithm for solar
    radiation applications. Solar Energy, vol. 76, no. 5, pp. 577-589, 2004.
    [5] I. Reda and A. Andreas, Corrigendum to Solar position algorithm for
    solar radiation applications. Solar Energy, vol. 81, no. 6, p. 838, 2007.
    [6] Blanc, P., & Wald, L., The SG2 algorithm for a fast and accurate
    computation of the position of the sun for multi-decadal time period, Solar
    Energy, 86(10), 3072–3083 (2012).

    The unfortunately quite computationally intensive SPA algorithm [4,5] has
    been implemented using numba or plain numpy for a single location at
    https://github.com/pvlib/pvlib-python/blob/master/pvlib/spa.py.

    """

    # up to h and dec from [1]

    t = ds.indexes['time']
    n = xr.DataArray(t.to_julian_date(), [ds.indexes['time']]) - 2451545.0

    L = 280.460 + 0.9856474 * n # mean longitude (deg)
    g = np.deg2rad(357.528 + 0.9856003 * n) # mean anomaly (rad)
    l = np.deg2rad(L + 1.915 * np.sin(g) + 0.020 * np.sin(2*g)) # ecliptic long. (rad)
    ep = np.deg2rad(23.439 - 4e-7 * n) # obliquity of the ecliptic (rad)

    ra = np.arctan2(np.cos(ep) * np.sin(l), np.cos(l)) # right ascencion (rad)
    lmst = (6.697375 + (ds['time.hour'] + ds['time.minute'] / 60.0) +
            0.0657098242 * n) * 15. + ds['lon'] # local mean sidereal time (deg)
    h = (np.deg2rad(lmst) - ra + np.pi) % (2*np.pi) - np.pi # hour angle (rad)

    dec = np.arcsin(np.sin(ep) * np.sin(l))            # declination (rad)

    # alt and az from [2]
    lat = np.deg2rad(ds['lat'])
    # Clip before arcsin to prevent values < -1. from rounding errors; can cause NaNs later
    alt = np.arcsin((np.sin(lat)*np.sin(dec) + np.cos(lat)*np.cos(dec)*np.cos(h)).clip(min=-1., max=1.)).rename('altitude')

    az = np.arccos(((np.sin(dec)*np.cos(lat) - np.cos(dec)*np.sin(lat)*np.cos(h))/np.cos(alt)).clip(min=-1., max=1.))
    az = az.where(h <= 0, 2*np.pi - az).rename('azimuth')

    if 'influx_toa' in ds:
        atmospheric_insolation = ds['influx_toa'].rename('atmospheric insolation')
    else:
        # [3]
        atmospheric_insolation = (1366.1 * (1+0.033*np.cos(g)) * np.sin(alt)).rename('atmospheric insolation')

    solar_position = xr.Dataset({da.name: da
                                 for da in [alt,
                                            az,
                                            atmospheric_insolation]})
    return solar_position
